package cn.ibizlab.core.data.model;

import cn.ibizlab.core.data.dto.BaseData;
import cn.ibizlab.core.data.dto.FilterData;
import cn.ibizlab.util.errors.BadRequestAlertException;
import cn.ibizlab.util.helper.DataObject;
import cn.ibizlab.util.helper.Inflector;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.alibaba.fastjson.annotation.JSONField;
import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.annotation.JsonInclude;
import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.Setter;
import lombok.experimental.Accessors;
import org.springframework.util.DigestUtils;
import org.springframework.util.ObjectUtils;
import org.springframework.util.StringUtils;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.math.BigDecimal;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.util.*;

@Getter
@Setter
@NoArgsConstructor
@Accessors(chain = true)
@JsonInclude(JsonInclude.Include.NON_EMPTY)
@JsonIgnoreProperties(ignoreUnknown = true)
public class PojoSchema {



    @JSONField(ordinal = 1)
    private String id;

    @JSONField(ordinal = 2)
    private String type;


    @JSONField(serialize = false)
    @JsonIgnore
    private String name;

    @JSONField(serialize = false)
    @JsonIgnore
    public String getName()
    {
        if(name==null)
            name=this.getOptions()==null?"":this.getOptions().getName();
        return name;
    }

    public PojoSchema setName(String name)
    {
        if(!StringUtils.isEmpty(name))
        {
            this.name=name;
            if(options==null)
                options=new PojoOption();
            options.setName(name);
        }
        return this;
    }

    @JSONField(serialize = false)
    @JsonIgnore
    public String getCodeName()
    {
        return this.getOptions()==null?this.getName():this.getOptions().getStringValue("code_name",this.getName());
    }

    @JSONField(serialize = false)
    @JsonIgnore
    public String getSystem()
    {
        return this.getOptions()==null?"":this.getOptions().getSystem();
    }

    @JSONField(serialize = false)
    @JsonIgnore
    public String getDefaultDataSoruce()
    {
        return this.getOptions()==null?"":this.getOptions().getDefaultDataSoruce();
    }

    @JSONField(serialize = false)
    @JsonIgnore
    public String getDefaultTableName()
    {
        return this.getOptions()==null?"":this.getOptions().getTableName();
    }

    @JSONField(serialize = false)
    @JsonIgnore
    public String getPackage()
    {
        return this.getOptions()==null?"":this.getOptions().getSystem();
    }



    @JSONField(ordinal = 3)
    private String title;

    @JSONField(ordinal = 4)
    private String description;

    @JSONField(ordinal = 5)
    private Integer propertyOrder;

    @JSONField(ordinal = 6)
    private PojoOption options;

    public PojoSchema setOptions(PojoOption options)
    {
        if(options==null)
            return this;

        if(this.options!=null)
            options.putAll(this.options);
        this.options=options;
        return this;
    }

    @JSONField(ordinal = 7)
    private String ref;

    @JSONField(serialize = false)
    @JsonIgnore
    public PojoSchema copy()
    {
        return JSONObject.parseObject(JSON.toJSONString(this),PojoSchema.class);
    }

    @JSONField(serialize = false)
    @JsonIgnore
    private boolean built = false;

    @JSONField(serialize = false)
    @JsonIgnore
    public PojoSchema build()
    {
        if((!built)&&Type.object.getCode().equalsIgnoreCase(this.getType()))
        {
            if(properties!=null)
            {
                properties.values().forEach(prop->{
                    PojoSchema item=null;
                    if(Type.object.getCode().equalsIgnoreCase(prop.getType())&&(!StringUtils.isEmpty(prop.getRef())))
                        item=prop;
                    else if(Type.array.getCode().equalsIgnoreCase(prop.getType())
                            &&prop.getItems()!=null&&Type.object.getCode().equalsIgnoreCase(prop.getItems().getType())
                            &&(!StringUtils.isEmpty(prop.getItems().getRef())))
                        item=prop.getItems();
                    else
                        return;
                });
            }
            built=true;
        }
        return this;
    }

    /// object

    @JSONField(ordinal = 11)
    private Map<String,PojoSchema> properties;

    public PojoSchema setProperties(Map<String,PojoSchema> properties)
    {
        if(properties!=null)
        {
            properties.keySet().forEach(key->{
                properties.get(key).setName(key).setOwner(this);
            });
            this.properties=properties;
        }
        return this;
    }

    public Map<String,PojoSchema> getProperties()
    {
        if(!Type.object.getCode().equals(this.type))
        {
            properties=null;
            return properties;
        }
        if(properties==null)
            properties = new LinkedHashMap<>();
        return properties;
    }

    public PojoSchema addProperty(String name,PojoSchema property)
    {
        if(StringUtils.isEmpty(name))
            return this;
        Map<String,PojoSchema> props = this.getProperties();
        if(props!=null) {
            props.put(name, property.setName(name).setOwner(this));
        }
        return this;
    }

    @JsonIgnore
    @JSONField(serialize = false)
    private Map<String,PojoSchema> references;

    @JsonIgnore
    @JSONField(serialize = false)
    public Map<String,PojoSchema> getReferences()
    {
        if(references==null)
        {
            references=new LinkedHashMap<>();
            getProperties().keySet().forEach(key->{
                PojoSchema prop=getProperties().get(key);
                if(Type.object.getCode().equalsIgnoreCase(prop.getType())&&prop.getOptions()!=null&&(!StringUtils.isEmpty(prop.getOptions().getRelationName()))) {
                    references.put(key, prop);
                    if(!key.equals(key.toLowerCase()))
                        references.put(key.toLowerCase(),prop);
                    references.put(prop.getOptions().getRelationName(),prop);
                    if(!StringUtils.isEmpty(prop.getOptions().getEntityName())) {
                        if(!references.containsKey(prop.getOptions().getEntityName()))
                            references.put(prop.getOptions().getEntityName(), prop);
                        if(!references.containsKey(prop.getOptions().getEntityName().toLowerCase()))
                            references.put(prop.getOptions().getEntityName().toLowerCase(),prop);
                    }
                    if(!StringUtils.isEmpty(prop.getOptions().getCodeName())) {
                        if(!references.containsKey(prop.getOptions().getCodeName()))
                            references.put(prop.getOptions().getCodeName(), prop);
                        if(!references.containsKey(prop.getOptions().getCodeName().toLowerCase()))
                            references.put(prop.getOptions().getCodeName().toLowerCase(), prop);
                    }
                    String pluralize=Inflector.getInstance().pluralize(StringUtils.isEmpty(prop.getOptions().getCodeName())?prop.getOptions().getEntityName():prop.getOptions().getCodeName()).toLowerCase();
                    if(!references.containsKey(pluralize))
                        references.put(pluralize, prop);
                }
            });
        }
        return references;
    }

    public PojoSchema getRefSchema(String tag)
    {
        if(getReferences()!=null)
        {
            return this.getReferences().get(tag);
        }
        return null;
    }

    public Map<String,PojoSchema> getRefProperties(String tag)
    {
        Map<String,PojoSchema> refProperties=new LinkedHashMap<>();
        PojoSchema refSchema=this.getRefSchema(tag);
        if(refSchema!=null)
        {
            getProperties().keySet().forEach(key-> {
                if(key.equals(refSchema.getName()))
                    return;
                PojoSchema prop = getProperties().get(key);
                if(prop.getOptions()!=null&&(!StringUtils.isEmpty(prop.getOptions().getRelationName()))&&prop.getOptions().getRelationName().equals(refSchema.getOptions().getRelationName()))
                {
                    refProperties.put(key,prop);
                }
            });
        }
        return refProperties;
    }


    @JSONField(serialize = false)
    @JsonIgnore
    private PojoSchema owner;

    @JSONField(ordinal = 12)
    private Set<String> required;

    public Set<String> getRequired()
    {
        if(!Type.object.getCode().equals(this.type))
        {
            required=null;
            return required;
        }
        if(required==null&&properties!=null)
        {
            required=new LinkedHashSet<>();
            properties.values().forEach(prop->{
                if (prop.getOptions()!=null&&(!prop.getOptions().isNullable()))
                    required.add(prop.getName());
            });
        }
        return required;
    }

    @JSONField(ordinal = 13)
    private Integer minProperties;

    @JSONField(ordinal = 14)
    private Integer maxProperties;


    @JsonIgnore
    @JSONField(serialize = false)
    private Map<String,PojoSchema> keyMap;

    @JsonIgnore
    @JSONField(serialize = false)
    private PojoSchema keyProperty;


    @JsonIgnore
    @JSONField(serialize = false)
    private Map<String,PojoSchema> unionKeys;

    public PojoSchema getKeyProperty() {
        if(keyProperty==null)
            getKeyMap();
        return keyProperty;
    }

    @JsonIgnore
    @JSONField(serialize = false)
    public synchronized Map<String,PojoSchema> getKeyMap()
    {
        if(Type.object.getCode().equals(this.type))
        {
            if(keyMap==null)
            {
                List<PojoSchema> keys=new ArrayList<>();
                keyMap=new LinkedHashMap<>();
                getProperties().values().forEach(sub->{
                    if(sub.getOptions().isKeyField())
                        keyProperty=sub;
                    if(sub.getOptions().isKeyField()&&sub.getOptions().isPhysicalField())
                        keys.add(sub);
                });
                if(keys.isEmpty())
                {
                    getProperties().values().forEach(sub->{
                        if(sub.getOptions().isUnionKeyField()&&sub.getOptions().isPhysicalField())
                            keys.add(sub);
                    });
                }
                if(keys.size()>0)
                    keys.sort( (o1, o2) -> o1.getOptions().getUnionKey().compareTo(o2.getOptions().getUnionKey()) );

                keys.forEach(sub->keyMap.put(sub.getOptions().getFieldName(),sub));
            }

        }
        return keyMap;
    }

    @JsonIgnore
    @JSONField(serialize = false)
    public synchronized Map<String,PojoSchema> getUnionKeys()
    {
        if(Type.object.getCode().equals(this.type)) {
            if (unionKeys == null) {
                List<PojoSchema> keys=new ArrayList<>();
                Map<String,PojoSchema> unions=new LinkedHashMap<>();

                if(keys.isEmpty())
                {
                    getProperties().values().forEach(sub->{
                        if(sub.getOptions().isUnionKeyField()&&sub.getOptions().isPhysicalField())
                            keys.add(sub);
                    });
                }
                if(keys.size()>0) {
                    keys.sort((o1, o2) -> o1.getOptions().getUnionKey().compareTo(o2.getOptions().getUnionKey()));
                    keys.forEach(sub -> unions.put(sub.getOptions().getFieldName(),sub));
                    unionKeys=unions;
                }
            }
        }
        return unionKeys;
    }


    ////  array

    @JSONField(ordinal = 21)
    private PojoSchema items;

    @JSONField(ordinal = 22)
    private Integer minItems;

    @JSONField(ordinal = 23)
    private Boolean uniqueItems;

    public void setUniqueItems(Boolean uniqueItems) {
        this.uniqueItems = uniqueItems;
    }

    ///  String

    @JSONField(ordinal = 33)
    private Integer minLength;

    @JSONField(ordinal = 34)
    private Integer maxLength;

    @JSONField(ordinal = 35)
    private String pattern;

    @JSONField(ordinal = 36)
    private String format;

    ///  integer number

    @JSONField(ordinal = 41)
    private BigDecimal minimum;

    @JSONField(ordinal = 42)
    private BigDecimal maximum;

    @JSONField(ordinal = 43)
    private Boolean exclusiveMinimum;

    @JSONField(ordinal = 44)
    private Boolean exclusiveMaximum;

    @JSONField(ordinal = 45)
    private Integer multipleOf;


    public PojoSchema writeTo(Path path)
    {
        return writeTo(this,path);
    }

    public static PojoSchema fromPath(Path path)
    {
        try {
            if(!Files.exists(path))
                throw new BadRequestAlertException("读取文件失败","PojoSchema",path.toString());
            return JSON.parseObject(Files.readAllBytes(path),PojoSchema.class);

        } catch (Exception e) {
            throw new BadRequestAlertException("解析文件失败","PojoSchema",path.toString());
        }
    }

    public static PojoSchema writeTo(PojoSchema pojoSchema,Path path)
    {
        try {
            File dir=path.getParent().toFile();
            if(!dir.exists())
                dir.mkdirs();

            Files.write(path, JSON.toJSONBytes(pojoSchema), StandardOpenOption.CREATE);
        } catch (Exception e) {
            throw new BadRequestAlertException("保存文件失败","PojoSchema",path.toString());
        }
        return pojoSchema;
    }


    public static enum Type {
        string("string", "字符"),
        integer("integer", "整型"),
        number("number", "数值"),
        object("object", "对象"),
        array("array", "数组");
        public final String code;
        public final String name;

        private Type(String code, String name) {
            this.code = code;
            this.name = name;
        }

        public String getCode() {
            return code;
        }

        public String getName() {
            return name;
        }

        @Override
        public String toString() {
            return "Type{" +
                    "code='" + code + '\'' +
                    ", name='" + name + '\'' +
                    '}';
        }
    }

    public Serializable getKeyValue(BaseData data, boolean genKeyWhenNotExists)
    {
        Object key=data.getKey();
        Map<String,PojoSchema> keyMap= this.getKeyMap();
        if(keyMap!=null)
        {
            PojoSchema keyProperty=this.getKeyProperty();
            if(ObjectUtils.isEmpty(key))
                key=data.get(keyProperty.getName().toLowerCase());

            if(ObjectUtils.isEmpty(key))
            {
                Map<String,PojoSchema> unionKeys=this.getUnionKeys();
                if(!ObjectUtils.isEmpty(unionKeys))
                {
                    for(PojoSchema keySchema:unionKeys.values())
                    {
                        if(ObjectUtils.isEmpty(data.get(keySchema.getName().toLowerCase()))) {
                            key=null;
                            break;
                        }
                        if(!StringUtils.isEmpty(key))
                            key+="||";
                        else
                            key+=data.getStringValue(keySchema.getName().toLowerCase());
                    }
                    if((!ObjectUtils.isEmpty(key))&&unionKeys.size()>1&&keyProperty.getOptions().isPhysicalField())
                        key= DigestUtils.md5DigestAsHex(key.toString().getBytes());
                }
            }
            if(ObjectUtils.isEmpty(key)&&genKeyWhenNotExists&&keyMap.size()==1)
            {
                if(PojoSchema.Type.integer.getCode().equals(keyProperty.getType())||PojoSchema.Type.number.getCode().equals(keyProperty.getType()))
                    key= IdWorker.getId();
                else
                    key=IdWorker.get32UUID();
            }
            if(ObjectUtils.isEmpty(key))
                return null;

            setKeyValue(data,key);

        }
        return (Serializable)key;
    }

    public BaseData newData(Object keyValue)
    {
        return setKeyValue(new BaseData(),keyValue);
    }

    public BaseData setKeyValue(BaseData data,Object keyValue)
    {
        Map<String,PojoSchema> keyMap= this.getKeyMap();
        data.setKey(keyValue);
        if(keyMap!=null)
        {
            PojoSchema keyProperty=this.getKeyProperty();
            data.set(keyProperty.getName().toLowerCase(),keyValue);
            Map<String,PojoSchema> unionKeys=this.getUnionKeys();
            if(!ObjectUtils.isEmpty(unionKeys))
            {
                String[] keys=DataObject.getStringValue(keyValue,"").split("\\|\\|");
                if(unionKeys.size()!=keys.length&&keyMap.size()>1)
                    return null;

                if(unionKeys.size()!=keys.length)
                {
                    int i=0;
                    for(PojoSchema keySchema:unionKeys.values())
                    {
                        if(ObjectUtils.isEmpty(data.get(keySchema.getName().toLowerCase()))) {
                            if(PojoSchema.Type.integer.getCode().equals(keySchema.getType()))
                                data.set(keySchema.getName().toLowerCase(),DataObject.getLongValue(keys[i],null));
                            else if(PojoSchema.Type.number.getCode().equals(keySchema.getType()))
                                data.set(keySchema.getName().toLowerCase(),DataObject.getBigDecimalValue(keys[i],null));
                            else if(PojoSchema.Type.string.getCode().equals(keySchema.getType())&&(!StringUtils.isEmpty(keySchema.getFormat()))&&keySchema.getFormat().indexOf("data")==0)
                                data.set(keySchema.getName().toLowerCase(),DataObject.getTimestampValue(keys[i],null));
                            else
                                data.set(keySchema.getName().toLowerCase(),keys[i]);
                        }
                        i++;
                    }
                }
            }
            return data;
        }

        return null;
    }

    public BaseData fillParentKey(BaseData data)
    {
        if(ObjectUtils.isEmpty(data.getParentData()))
        {
            return data;
        }
        data.getParentData().entrySet().forEach(entry->{
            String refSchema=entry.getKey();
            BaseData parent=entry.getValue();
            for(Map.Entry<String,PojoSchema> keyset:this.getRefProperties(refSchema).entrySet())
            {
                PojoSchema prop=keyset.getValue();
                if(prop.getOptions()!=null)
                {
                    if("PICKUP".equalsIgnoreCase(prop.getOptions().getFieldType())) {
                        data.set(keyset.getKey().toLowerCase(), parent.getKey());
                        break;
                    }
                }
            }

        });
        return data;
    }

    public FilterData fillParentFilter(FilterData data)
    {
        if(ObjectUtils.isEmpty(data.getParentData()))
        {
            return data;
        }
        data.getParentData().entrySet().forEach(entry->{
            String refSchema=entry.getKey();
            BaseData parent=entry.getValue();
            for(Map.Entry<String,PojoSchema> keyset:this.getRefProperties(refSchema).entrySet())
            {
                PojoSchema prop=keyset.getValue();
                if(prop.getOptions()!=null)
                {
                    if("PICKUP".equalsIgnoreCase(prop.getOptions().getFieldType())) {
                        if(!StringUtils.isEmpty(prop.getOptions().getFieldName()))
                            data.set("n_"+prop.getOptions().getFieldName()+"_eq", parent.getKey());
                        break;
                    }
                }
            }

        });
        return data;
    }

}
